/*****************************************************************
 * FILE: alignalign.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: 17 May 2007
 * Version:  1.0.0
 * 
 ****************************************************************/

#include <stdlib.h> /* NULL */
#include <stdio.h> /* fprintf */
#include <string.h> /* strlen */
#include "alignalign.h"


/* global variables */
char  **aa_A = NULL;                      /* storage for input alignment aa_A     */
char  **aa_B = NULL;                      /* storage for input alignment aa_B     */
int **aa_cost;/* aa_cost[a,b] = substitution cost     */
int     aa_gamma          = AA_GAMMA;  /* gap open penalty                  */
int     aa_lambda         = AA_LAMBDA; /* gap extension penalty             */
int     aa_gamma_terminal   = AA_GAMMA;  /* gap open penalty                  */
int     aa_lambda_terminal  = AA_LAMBDA; /* gap extension penalty             */

int     aa_weighted          = AA_FALSE;  /* command line option flags         */
int **aa_weights;/* aa_weights[i,j] = weight of sequence pair (i,j)     */
int aa_weightMatrixSize;
int     aa_printseqs         = AA_FALSE;  /* */
int     aa_printcost         = AA_FALSE;  /* */
int     aa_printinput        = AA_FALSE;  /* */
int     aa_M = 0, aa_N = 0, aa_K = 0, aa_L = 0;    /* aa_M=|aa_A|,aa_N=|aa_B|,aa_K=# aa_A seqs,aa_L=# aa_B seqs */
int    *aa_firstCharsA;
int    *aa_firstCharsB;
int    *aa_lastCharsA;
int    *aa_lastCharsB;
int     aa_nodesize = 0;
int     aa_nodecount = 0;
void   *aa_nodeptr = NULL;
Shape  *aa_nodepool = NULL;
void   *aa_nodepoolroot = NULL;
int     aa_nodepoolcount = 0;
int     aa_nodemalloccount = 0;
int     aa_nodemallocsize = 0;
aa_ulong aa_ec1 = 0;
aa_uchar **aa_fa; // For profile of columns - for linear-time 
aa_uchar **aa_fb;
aa_uchar **aa_ca;
aa_uchar **aa_cb;
int  *aa_alphaconv;
char *aa_alphabet;
int  aa_alphasize;
int   aa_cost_upper_bound = 0;
int **aa_optH = NULL, **aa_optV = NULL, **aa_optD = NULL; //TW - tables for bound pruning

unsigned char *aa_script;

void 
aa_setCosts ( int ***cost, int alpha_size, char *alphabet, 
			int *alphaconv,	int gamma, int lambda, int gamma_terminal, int lambda_terminal)
{    	
  // copy input vars into alignalign globals
  aa_gamma = gamma;
  aa_lambda = lambda;
  aa_gamma_terminal = gamma_terminal;
  aa_lambda_terminal = lambda_terminal;
  aa_alphasize = alpha_size;
  aa_alphaconv = alphaconv;
  aa_alphabet = alphabet;  
  aa_cost = *cost;				  
}

void 
aa_setWeights ( int K, int **weights)
{    
	
//	printf ("setting weights:\n");
    aa_freeWeights();
    int i,j;
    // copy input var into alignalign globals	
	aa_weights = aa_malloc( K*sizeof(int*));
	for (i=0; i<K; i++) {
		aa_weights[i] = aa_malloc(K*sizeof(int));
		for (j=0; j<K; j++) {
			aa_weights[i][j] = (NULL==weights)?1.0:weights[i][j]; 
//			printf("%d\t",aa_weights[i][j]);
		}
//		printf ("\n");
	}
    aa_weightMatrixSize = K;
    if ( NULL != weights) 
	    aa_weighted = AA_TRUE;    

}

void aa_setUnweighted() {
	aa_freeWeights();
    aa_weightMatrixSize = -1;
    aa_weighted = AA_FALSE;
}


void make_defaultWeights(int K) {
	aa_setWeights(K, NULL);
}

void 
aa_freeWeights ()
{    
  if (!aa_weighted || NULL == aa_weights){
  	return;
  }
  int i;
  // copy input var into alignalign globals
  for (i=0; i<aa_weightMatrixSize; i++) 
  	if (NULL != aa_weights[i])
	  	free(aa_weights[i]);

  aa_weightMatrixSize = -1;
  free(aa_weights);
}

	

int 
alignalign(	int space_optimized, char ***A, int K, int M, 
	char ***B, int L, int N, char ***alignment_ptr, unsigned char **script_ptr)
{
	  
	
  // copy input vars into alignalign globals
  aa_K = K;
  aa_L = L;
  aa_A = *A;
  aa_B = *B;
  //done copying to aa_globals  

  int alignlen = 0;
  char   **temp_charpp = NULL;            /* temporary varible        */
  int      temp_int = 0;                  /* temporary varible        */
  int alignmentsSwapped = AA_FALSE;
   
  if (!aa_weighted) {
  	make_defaultWeights(K+L);
  }
  init_nodepool();    /* initialize pool of Shapes */



  /* remove null columns from input alignments */
  aa_M = RemoveNullColumns(aa_A, K, M);		  
  aa_N = RemoveNullColumns(aa_B, L, N);


  int i,j;

/*printf ("I've been asked to align the following:\n");
for (i=0; i<K; i++) {
	for (j=0; j<aa_M; j++) {
		printf ("%c",aa_A[i][j]);
	}
	printf ("\n");
}
printf ("versus\n");
for (i=0; i<L; i++) {
	for (j=0; j<aa_N; j++) {
		printf ("%c",aa_B[i][j]);
	}
	printf ("\n\n");
}
fflush(stdout);
*/
  FindFirstAndLastChars (aa_A, &aa_firstCharsA, &aa_lastCharsA, aa_K, aa_M ) ;
  FindFirstAndLastChars (aa_B, &aa_firstCharsB, &aa_lastCharsB, aa_L, aa_N ) ;    


  if (!(aa_weighted || aa_gamma_terminal != aa_gamma || aa_lambda_terminal != aa_lambda)) {
	  init_frequencies(); /* initialize frequency profile */
  }
  
  if (space_optimized) {
	 /* swap aa_A & aa_B if |aa_A| < |aa_B| */
	  if (aa_M < aa_N) {
	    temp_int = aa_K; aa_K = aa_L; aa_L = temp_int;
	    temp_int = aa_M; aa_M = aa_N; aa_N = temp_int; 
	    temp_charpp = aa_A; aa_A = aa_B; aa_B = temp_charpp;
		alignmentsSwapped = AA_TRUE;
		
		//swap weight matrix around
		int** new_weights = aa_malloc( (K+L)*sizeof(int*));
		for (i=0; i<K+L; i++) 
			new_weights[i] = aa_malloc((K+L)*sizeof(int));
	
		for (i=0; i<K; i++) // move upper-left quadrant to lower-right
			for (j=0; j<K; j++) 
				new_weights[L+i][L+j] = aa_weights[i][j] ;
		for (i=0; i<L; i++) // move lower-right quadrant to upper-left
			for (j=0; j<L; j++) 
				new_weights[i][j] = aa_weights[K+i][K+j] ;
		for (i=0; i<L; i++) // move lower-left quadrant to upper-right
			for (j=0; j<K; j++) 
				new_weights[i][L+j] = aa_weights[K+i][j] ;
		for (i=0; i<K; i++) // move upper-right quadrant to lower-left	
			for (j=0; j<L; j++) 
				new_weights[K+i][j] = aa_weights[i][L+j] ;
		aa_setWeights(K, new_weights);
		for (i=0; i<K+L; i++) 
			free(new_weights[i]);
		free(new_weights);
		
	  }


	alignlen = alignalign_space(alignment_ptr);
  }
  else {
    alignlen = alignalign_time(alignment_ptr);
  } 

/*  If alignments were swapped, swap them back again, so they're in 
 * 		the same order in which they were passed-in.
  	  This means the first L seqs should actually be the last L seqs.
  	  Create L temporary pointers, move those entries out, shift up the
  	  remaining entries, then put the L entries where they should be
  */	  

  if (alignmentsSwapped) {
  	  char ** tmp_alignment;
  	  
      tmp_alignment = aa_malloc((K+L) * sizeof(char *));
  	  memcpy (tmp_alignment, *alignment_ptr, (K+L) * sizeof(char *));

  	  //move other alignment up.
  	  for (i = 0; i < K; i++) {
  	  	(*alignment_ptr)[i] = tmp_alignment[L+i];
  	  }  	
  	  //move tmp back in
  	  for (i = 0; i < L; i++) {
  	  	(*alignment_ptr)[K+i] = tmp_alignment[i];
  	  }  	
	  free(tmp_alignment);
  }
  
  if (script_ptr != NULL) {
	if (alignmentsSwapped) {
		//insertion and deletion mean the opposite thing once we swap the splits
		for (i=0; i<alignlen; i++){
			if (aa_script[i] == AA_DEL)
				aa_script[i] = AA_INS;
			else if (aa_script[i] == AA_INS)
				aa_script[i] = AA_DEL;
		}
	}
 	*script_ptr = aa_script; //calling app is responsible for freeing this
  } else {
	free(aa_script);
  }
  
  if (!(aa_weighted || aa_gamma_terminal != aa_gamma || aa_lambda_terminal != aa_lambda)) {
	  free_frequencies();
  }
   
  free (aa_firstCharsA);
  free (aa_firstCharsB);
  free (aa_lastCharsA);
  free (aa_lastCharsB);
   
  return alignlen;
}



/********************************************************************
 * Procedure: RemoveSpacerColumns  (remove spacer columns from MSA M)
 *******************************************************************/
int RemoveNullColumns (char **M, int rows, int cols) {
  int i, j, k, flag;
  
  i = 0; 
  flag = AA_FALSE;
  for (j = 0; j < cols; j++) {
    for (k = 0; k < rows; k++) {
      M[k][i] = M[k][j];
      if (M[k][j] != AA_EPS)
      	flag = AA_TRUE; 
  	}
    if (flag == AA_TRUE) { i++; flag = AA_FALSE; } 
  }
  if (i < cols) {
	  for (j = 0; j < rows; j++) { M[j][i] = '\0'; }
  }
  return i; /* new number of columns */
}

void FindFirstAndLastChars ( char **M, int **firstChars, int **lastChars, int rows, int cols ) {
  int k,j;

  *firstChars = aa_malloc ( rows * sizeof(int));
  *lastChars = aa_malloc ( rows * sizeof(int));
  
  for (k = 0; k < rows; k++) {
    j=0;
    while (j < cols) {
    	if (AA_EPS != M[k][j]){
    		(*firstChars)[k] = j;
    		break;	
    	}
    	j++;
    }

    j=cols-1;
    while (j >= 0) {
    	if (AA_EPS != M[k][j]){
    		(*lastChars)[k] = j;
    		break;	
    	}
    	j--;
    }
    if (j == 0 && AA_EPS==M[k][0]) {
	    fprintf(stderr, "\t sequence in input that contains only dashes\n");
    }
  }
}


/* if the position "col" in the string at "row" comes outside the first and last non-gap 
 * characters, use the "terminal" gap costs */
inline int GetGamma (int *first, int *last, int row, int col) {
	if (col < first[row]) {
		return aa_gamma_terminal;
	} else if ( col >= last[row] ) {
		return aa_gamma_terminal;
	} else {
		return aa_gamma;		
	}
	//return ( col < first[row] || col >= last[row] ) ? aa_gamma_terminal : aa_gamma;
}

/* if the position "col" in the string at "row" comes outside the first and last non-gap 
 * characters, use the "terminal" gap costs */
inline int GetLambda (int *first, int *last, int row, int col) {
	return ( col < first[row] || col >= last[row] ) ? aa_lambda_terminal : aa_lambda;
}


/* Because one step of the alignalign_time requires filling the DP table in reverse,
 * we need to reverse the values in the "firstchar" and "lastchar" arrays ... then flip them back.
 * This function does that */
void flipFirstLastCharArrays() {
	int tmp, i;
	
	for (i=0; i<aa_K; i++) {
		tmp = aa_firstCharsA[i];
		aa_firstCharsA[i] = aa_M - 1 - aa_lastCharsA[i];
		aa_lastCharsA[i] = aa_M - 1 - tmp;
	}
	for (i=0; i<aa_L; i++) {
		tmp = aa_firstCharsB[i];
		aa_firstCharsB[i] = aa_N - 1 - aa_lastCharsB[i];
		aa_lastCharsB[i] = aa_N - 1 - tmp;
	}
}
