/*****************************************************************
 * FILE: dominance.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: 17 May 2007
 * Version:  1.0.0
 * 
 ****************************************************************/

#include <stdlib.h>  /* NULL */
#include "bound.h"
#include "alignalign.h"

/*
 * procedure: dom_gapbound_weighted
 *   Input:  shapes s and t
 *   Output: gc, a int representing an upperbound on the cost of
 *           gaps that could possibly be started in s but not in t,
 *           over all possible extensions.
 *  This is a O(aa_K * aa_L) version, slower than "unweighted", but handles sequence pair weights
 */
inline int
dom_gapbound_weighted (aa_ushort *s, aa_ushort *t, int a, int b) {
  int gc = 0.0;
  int i,j;
  for ( i=0; i<aa_K; i++) {
  	for ( j=aa_K; j<aa_K+aa_L; j++) {
  		if 	( t[i]>t[j] && s[i]<=s[j] ) {
  			gc += aa_weights[i][j] * GET_GAMMA_B(j-aa_K,b);
  		} else if ( t[i]<t[j] && s[i]>=s[j] ) {
  			gc += aa_weights[i][j] * GET_GAMMA_A(i,a);	
  		}
  	}
  }
  return gc;
}

/*
 * procedure: dom_gapbound_unweighted
 *   Input:  shapes s and t
 *   Output: gc, a int representing an upperbound on the cost of
 *           gaps that could possibly be started in s but not in t,
 *           over all possible extensions.
 *  This is a O(aa_K + aa_L) version based on the ideas in: O. Gotoh, "Further 
 *  improvements in methods of group-togroup sequence alignment with 
 *  generalized profile operations. CABIOS 10(4):379-387, 1994.
 */
inline int
dom_gapbound_unweighted (aa_ushort *s, aa_ushort *t) {
  static int gc;
  static int i, j, os, ot, us, ut;
  static int sablocknum, tablocknum, sbblocknum, tbblocknum;
  static int sablock, tablock, sbblock, tbblock;
  static int Bas[AA_MAXSEQNUM];
  static int Bat[AA_MAXSEQNUM];
  static int Bbs[AA_MAXSEQNUM];
  static int Bbt[AA_MAXSEQNUM];
  static int Basc[AA_MAXSEQNUM];
  static int Batc[AA_MAXSEQNUM];
  static int Bbsc[AA_MAXSEQNUM];
  static int Bbtc[AA_MAXSEQNUM];
  static int Basb[AA_MAXSEQNUM];
  static int Batb[AA_MAXSEQNUM];
  static int Bbsb[AA_MAXSEQNUM];
  static int Bbtb[AA_MAXSEQNUM];
  static int Bascp[AA_MAXSEQNUM];
  static int Batcp[AA_MAXSEQNUM];
  static int Bbscp[AA_MAXSEQNUM];
  static int Bbtcp[AA_MAXSEQNUM];
  static int Oas[AA_MAXSEQNUM];
  static int Oat[AA_MAXSEQNUM];
  static int Uas[AA_MAXSEQNUM];
  static int Uat[AA_MAXSEQNUM];


  /*precomputation */
  //number bocks in the A-portion ("[st]ablocknum") (or B-portion for "[stbblocknum]"
  // of the block-array for shapes s and t
  sablocknum = 0;
  tablocknum = 0;
  sbblocknum = 0;
  tbblocknum = 0;  
  for (i = 0; i < aa_K; i++) {
    if (s[i] > sablocknum) { sablocknum = s[i]; }
    if (t[i] > tablocknum) { tablocknum = t[i]; }
  }
  for (i = aa_K; i < aa_K + aa_L; i++) {
    if (s[i] > sbblocknum) { sbblocknum = s[i]; }
    if (t[i] > tbblocknum) { tbblocknum = t[i]; }
  }
  sablocknum++;
  tablocknum++;  
  sbblocknum++;
  tbblocknum++;
  
  //initialize
  for (i = 0; i < sablocknum; i++) {
    Bas[i] = 0;
    Bascp[i] = 0;
  }
  for (i = 0; i < tablocknum; i++) {
    Bat[i] = 0;
    Batcp[i] = 0;
  }
  for (i = 0; i < sbblocknum; i++) {
    Bbs[i] = 0;
    Bbscp[i] = 0;
  }
  for (i = 0; i < tbblocknum; i++) {
    Bbt[i] = 0;
    Bbtcp[i] = 0;
  }

  //number of seqs in the A-sequence-set of each block for shapes s and t
  for (i = 0; i < aa_K; i++) {
    Bas[s[i]]++;
    Bat[t[i]]++;
  }
  //number of seqs in the B-sequence-set of each block for shapes s and t
  for (i = aa_K; i < aa_K + aa_L; i++) {
    Bbs[s[i]]++;
    Bbt[t[i]]++;
  }

  //compress the arrays of # per block
  j = 0;
  for (i = 0; i < sablocknum; i++) {
    if (0 < Bas[i]) {
      Basc[j] = Bas[i];
      Basb[j] = i;
      j++;
    }
  }
  sablock = j;
  j = 0;
  for (i = 0; i < sbblocknum; i++) {
    if (0 < Bbs[i]) {
      Bbsc[j] = Bbs[i];
      Bbsb[j] = i;
      j++;
    }
  }
  sbblock = j;
  j = 0;
  for (i = 0; i < tablocknum; i++) {
    if (0 < Bat[i]) {
      Batc[j] = Bat[i];
      Batb[j] = i;
      j++;
    }
  }
  tablock = j;
  j = 0;
  for (i = 0; i < tbblocknum; i++) {
    if (0 < Bbt[i]) {
      Bbtc[j] = Bbt[i];
      Bbtb[j] = i;
      j++;
    }
  }
  tbblock = j;

  //At this point, B[ab][st]c[i] contains the number of sequences in sequence-set (a,b) of shape (s,t) 
  // found in block number B[ab]b[i]


  //cumulative counts
  Bascp[0] = Basc[0];
  Batcp[0] = Batc[0];
  for (i = 1; i < sablock; i++) {
    Bascp[i] = Bascp[i-1] + Basc[i];
    Batcp[i] = Batcp[i-1] + Batc[i];
  }
  Bbscp[0] = Bbsc[0];
  Bbtcp[0] = Bbtc[0];
  for (i = 1; i < sbblock; i++) {
    Bbscp[i] = Bbscp[i-1] + Bbsc[i];
    Bbtcp[i] = Bbtcp[i-1] + Bbtc[i];
  }
  //At this point, B[ab][st]cp[i] contains the number of sequences in sequence-set (a,b) of shape (s,t) 
  // found *in-or-before* block B[ab]b[i]

  os = ot = 0;
  for (i = 0; i < sablock; i++) {
    while (Basb[i] > Bbsb[os] && os < sbblock) { os++; } //shape s: how many block groups from sequence-set b are to the left of the ith block in sequence-set a? os. 
    while (Batb[i] > Bbtb[ot] && ot < sbblock) { ot++; } //same for shape t.
    Oas[i] = os - 1;
    Oat[i] = ot - 1;
  }
  us = ut = 0;
  for (i = 0; i < sbblock; i++) {
    while (Basb[us] < Bbsb[i] && us < sablock) { us++; }  // same as above, but swap "a" and "b" sequence-sets
    while (Batb[ut] < Bbtb[i] && ut < sablock) { ut++; }
    Uas[i] = us - 1;
    Uat[i] = ut - 1;
  }

  //Here, Oas[i] contains the number of blocks in sequence-set b that are to the left of block i in sequence-set a, for shape s.
  // others ([UO][ab][st]) are the same.


  gc = 0.0;
  for (i = 0; i < sablock; i++) {
    if (Oas[i] < Oat[i]) {
      if (-1 == Oas[i]) {
		gc += Basc[i] * Bbtcp[Oat[i]];   //TW: should be Basc[i] * Bbscp[Oat[i]] ?
      }
      else {
		gc += Basc[i] * (Bbtcp[Oat[i]] - Bbscp[Oas[i]]);
      }
    }
  }

  for (i = 0; i < sbblock; i++) {
    if (Uas[i] < Uat[i]) {
      if (-1 == Uas[i]) {
		gc += Bbsc[i] * Batcp[Uat[i]];
      }
      else {
		gc += Bbsc[i] * (Batcp[Uat[i]] - Bascp[Uas[i]]);
      }
    }
  }

  gc *= aa_gamma;

  return gc;
}


/*
 * procedure: dom_gapbound_linear
 *   Input:  shapes s and t
 *   Output: gc, an integer representing an upperbound on the cost of
 *           gaps that could possibly be started in t but not in s,
 *           over all possible extensions.
*/
inline int
dom_gapbound (aa_ushort *s, aa_ushort *t, int a, int b) {
	if (aa_weighted || aa_gamma != aa_gamma_terminal  || aa_lambda != aa_lambda_terminal ) {	
		return dom_gapbound_weighted(s,t,a,b);
	} else {
		return dom_gapbound_unweighted(s,t);
	}
}

/*
 * Given shape s with associated cost c and a list of shapes, 
 * dominance returns true if s is dominated by some shape in the 
 * list and false otherwise.  All shapes in list that are dominated 
 * by s are removed from list.
 */
inline int 
dominance(Shape *list, aa_ushort *s, int c, int a, int b) {
  int     bound_on_gap_difference;
  Shape  *p = NULL, *q = NULL;
  
  /* if (NULL == list) { return AA_FALSE; } */

  /* check to see if s is dominated by a shape already in the list */
  q = list;
  p = NEXT(list);

  while (p != list) {
    /* check for current shape from list dominating s */
	if ( C(p) > c ) { 
		//no way S(p) can dominate s, so don't waste time running dom_gapbound
	} else {
		bound_on_gap_difference = dom_gapbound(S(p), s, a, b);
    	if (c >= C(p) + bound_on_gap_difference) { return AA_TRUE; }
	}
	
    /* check for s dominating current shape from list */
    if ( c > C(p) ) { 
    	//no way s can dominate S(p), so don't waste time running dom_gapbound
    	q = p; p = NEXT(p);
    } else {
	    bound_on_gap_difference = dom_gapbound(s, S(p), a, b);	
	    if (C(p) >= c + bound_on_gap_difference) {
	      NEXT(q) = NEXT(p);
	
	      /* current shape is dominated by s, so remove it from list */
	      nodefree(p);
	      p = NEXT(q);
	
	      /* s dominates another, so nothing dominates it (dom is transitive) */
	      break;
	    } else {     /* advance to next shape in list */
	    	q = p; p = NEXT(p); 
	    }
    }
    

  }
  
  //TW - broke out to get here.  s dominated current_s, so don't need to check if s is dominated ... but still want to see if it dominates anything else
  /* check to see if s dominates other shapes in the list */
  while (p != list) {
	if ( c > C(p) ) { 
		//no way s can dominate S(p), so don't waste time running dom_gapbound
		q = p; p = NEXT(p);
	} else {
	    bound_on_gap_difference = dom_gapbound(s, S(p), a, b);
	    if (C(p) >= c + bound_on_gap_difference) {
	      NEXT(q) = NEXT(p);
	      nodefree(p);
	      p = NEXT(q);
	    }
	    else { q = p; p = NEXT(p); }
	}
  }
  return AA_FALSE;
}
