/*****************************************************************
 * FILE: gaps.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: 17 May 2007
 * Version:  1.0.0
 * 
 ****************************************************************/

#include <stdlib.h> /* NULL */
#include <stdio.h>  /* print */
#include <string.h> /* memset */
#include "alignalign.h"


/*
 * procedure: gaps_unweighted
 *   Input:
 *     a, b: column indices into aa_A and aa_B respectively
 *        s: integer array representation of a shape
 *   Output: gc, a int representing the sum of pairwise
 *           gap startup penalties between aa_A and aa_B (not within aa_A, nor aa_B)
 *           resulting from appending column a of aa_A and column b of aa_B 
 *           to shape s (a < 0 indicates deletion, b < 0 indicates insertion).
 *  This is a O(aa_K + aa_L) version. For details see O. Gotoh, "Further 
 *  improvements in methods of group-togroup sequence alignment with 
 *  generalized profile operations. CABIOS 10(4):379-387, 1994.
 */
inline int
gaps_unweighted (int a, int b, aa_ushort *s, int dir) {
  static int gc;
  static int i, blocknum;
  static int Ta[AA_MAXSEQNUM]; //Ta[i] := num gap chars in block i of alignment A
  static int Tb[AA_MAXSEQNUM]; //Tb[i] := num gap chars in block i of alignment B
  static int Sa[AA_MAXSEQNUM]; //Sa[i] := (1) num non-gap chars in block i of alignment A
 								//   then (2) num strings whose final non-gap char occurs at or before block i in A.
  static int Sb[AA_MAXSEQNUM]; //Sb[i] := same Sa, but for B.

  blocknum = 0;
  for (i = 0; i < aa_K + aa_L; i++) {
    if (s[i] > blocknum) { blocknum = s[i]; }
  }
  blocknum++;

  for (i = 0; i < blocknum; i++) {
    Ta[i] = 0;
    Tb[i] = 0;
    Sa[i] = 0;
    Sb[i] = 0;
  }
  //memset(Ta, 0, blocknum * sizeof(int));
  //memset(Tb, 0, blocknum * sizeof(int));
  //memset(Sa, 0, blocknum * sizeof(int));
  //memset(Sb, 0, blocknum * sizeof(int));

  gc = 0;
  /* aligning column from aa_A with a null column */
  if (/*0 > b*/ AA_VERT == dir) {
    for (i = 0; i < aa_K; i++) {
      if (AA_EPS == aa_A[i][a]) { (Ta[s[i]])++; }
      else                { (Sa[s[i]])++; }
    }

    for (i = 0; i < aa_L; i++) {
       (Tb[s[i+aa_K]])++;
    }
    for (i = 1; i < blocknum; i++) { 
      Sa[i] += Sa[i-1];
    }
  }

  /* aligning column from aa_B with a null column */
  else if (/*0 > a*/ AA_HORIZ == dir) {
    for (i = 0; i < aa_K; i++) {
       (Ta[s[i]])++;
    }
    for (i = 0; i < aa_L; i++) {
      if (AA_EPS == aa_B[i][b]) { (Tb[s[i+aa_K]])++; }
      else                { (Sb[s[i+aa_K]])++; }
    }
    for (i = 1; i < blocknum; i++) { 
      Sb[i] += Sb[i-1];
    }
  }

  /* aligning a column from aa_A with a coumn from aa_B */
  else {
    for (i = 0; i < aa_K; i++) {
      if (AA_EPS == aa_A[i][a]) { (Ta[s[i]])++; }
      else                { (Sa[s[i]])++; }
    }
    for (i = 0; i < aa_L; i++) {
      if (AA_EPS == aa_B[i][b]) { (Tb[s[i+aa_K]])++; }
      else                { (Sb[s[i+aa_K]])++; }
    }
    for (i = 1; i < blocknum; i++) { 
      Sa[i] += Sa[i-1];
      Sb[i] += Sb[i-1];
    }
  }

  /**/
  for (i = 0; i < blocknum; i++) {
    gc += (Ta[i] * Sb[i] + Tb[i] * Sa[i]);
  }
  return gc * aa_gamma;
}


/*
 * procedure: gaps_weighted
 *   Input:
 *     a, b: column indices into aa_A and aa_B respectively
 *        s: integer array representation of a shape
 *   Output: gc, a int representing the sum of pairwise
 *           gap startup penalties between aa_A and aa_B (not within aa_A, nor aa_B)
 *           resulting from appending column a of aa_A and column b of aa_B 
 *           to shape s (a < 0 indicates deletion, b < 0 indicates insertion).
 *  This is a O(aa_K * aa_L) version, slower than "unweighted" because it handled p/w weights 
 */
inline int
gaps_weighted (int a, int b, aa_ushort *s, int dir) {
  int gc=0.0;
  int i,j;

  for (i = 0; i < aa_K; i++) {
    for (j = 0; j < aa_L; j++) {
 	  	/*if (0 > b) {  aligning column from aa_A with a null column */
 		if (/*0 > b*/ AA_VERT == dir) {
  				if (AA_EPS != aa_A[i][a] && s[i] <= s[aa_K+j] /*B[j] doesn't underhang A[i]*/) { 
				gc += aa_weights[i][aa_K+j] * GET_GAMMA_B(j,b-1);
			}
/*	    } else if (0 > a) {   aligning column from aa_B with a null column */
 		} else if (/*0 > a*/ AA_HORIZ == dir) {
			if (AA_EPS != aa_B[j][b] && s[i] >= s[aa_K+j] /*A[i] doesn't underhang B[j]*/) { 
				gc += aa_weights[i][aa_K+j] * GET_GAMMA_A(i,a-1);
			}
 	  	}  else {    /* aligning a column from aa_A with a coumn from aa_B */
			if ( AA_EPS == aa_A[i][a] && AA_EPS != aa_B[j][b] && s[i] >= s[aa_K+j] ) {
				gc += aa_weights[i][aa_K+j] * GET_GAMMA_A(i,a);				
			} else if (AA_EPS != aa_A[i][a] && AA_EPS == aa_B[j][b] && s[i] <= s[aa_K+j] ) { 
				gc += aa_weights[i][aa_K+j] * GET_GAMMA_B(j,b);
			}
 	  	}
    }
  }
  return gc;
}


/*
 * procedure: gaps
 *   Input:
 *     a, b: column indices into aa_A and aa_B respectively
 *        s: integer array representation of a shape
 *   Output: gc, a int representing the sum of pairwise
 *           gap startup penalties between aa_A and aa_B (not within aa_A, nor aa_B)
 *           resulting from appending column a of aa_A and column b of aa_B 
 *           to shape s (a < 0 indicates deletion, b < 0 indicates insertion).
 */
int
gaps (int a, int b, aa_ushort *s, int dir) {
	if (aa_weighted || aa_gamma != aa_gamma_terminal  || aa_lambda != aa_lambda_terminal) {
		return gaps_weighted(a,b,s,dir);
	} else {
		return gaps_unweighted(a,b,s,dir);		
	}
}
