/*****************************************************************
 * FILE: pessimistic_optimistic.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: 17 May 2007
 * Version:  1.0.0
 * 
 ****************************************************************/

//#include <stdio.h>		/* fprintf */
#include "pessimistic_optimistic.h"
#include "alignalign.h"

/**********************************************************************
 * procedure: pessimistic_optimistic
 *   Input:
 *     aa_A, aa_B: multiple sequence alignments
 *     k, l: number of rows in aa_A and aa_B respectively
 *     m, n: number of columns in aa_A and aa_B respectively
 *        d: substitution aa_cost matrix
 *        r: real value in the range [0,1], r*aa_gamma is charged for a 
 *           gap in the case where itcan't be determined if a gap opens.
 *   Output: qc, an integer representing the actual aa_cost of the optimal
 *           hueristic alignment of aa_A and aa_B using pessimistic or optimistc
 *           (or between) gap counts (optimistic if r = 0, pessimistic if
 *           r = 1, and whatever you want to call it if 0 < r < 1.
 *           
 *********************************************************************/
int 
pessimistic_optimistic_incl_alignment (char **A, char **B, 
			int k, int l, int m, int n, double r, char ***return_alignment, 
			int *return_len, unsigned char **script ) {

	if (aa_weighted || aa_gamma != aa_gamma_terminal  || aa_lambda != aa_lambda_terminal ) {
		return pessimistic_optimistic_incl_alignment_weighted (A, B, k, l, m, n, r, return_alignment, return_len, script);
	} else {
		return pessimistic_optimistic_incl_alignment_unweighted (A, B, k, l, m, n, r, return_alignment, return_len, script);
	}
}

/**********************************************************************
 * procedure: pessimistic_optimistic_unweighted
 *   Input:
 *     aa_A, aa_B: multiple sequence alignments
 *     k, l: number of rows in aa_A and aa_B respectively
 *     m, n: number of columns in aa_A and aa_B respectively
 *        d: substitution aa_cost matrix
 *        r: real value in the range [0,1], r*aa_gamma is charged for a 
 *           gap in the case where itcan't be determined if a gap opens.
 *   Output: qc, a int representing the actual aa_cost of the optimal
 *           hueristic *unweighted* alignment of aa_A and aa_B using pessimistic or optimistc
 *           (or between) gap counts (optimistic if r = 0, pessimistic if
 *           r = 1, and whatever you want to call it if 0 < r < 1.
 *           
 *********************************************************************/
int
pessimistic_optimistic_incl_alignment_unweighted (char **A, char **B, 
			int k, int l, int m, int n, double r, char ***return_alignment, 
			int *return_len, unsigned char **script ) {

  int     i, j, x, y, flag, len;
  int   a, b, c, qc, min, sub;
  int    *f_1A, *f_1B, *f_firstA, *f_firstB, *gapstateH,*gapstateV, *gapstateD;
  int    *f_10A, *f_10B, *f_01A, *f_01B, *f_00A, *f_00B, *f_0A, *f_0B;
  int **H, **V, **D;
  char  **al;

    
  f_0A      = aa_malloc((m+1) * sizeof(int)); //f_0A[j] = number of gap chars at position A[j-1]
  f_0B      = aa_malloc((n+1) * sizeof(int)); //f_0B[j] = number of gap chars at position B[j-1]
  f_1A      = aa_malloc((m+1) * sizeof(int)); //f_1A[j] = number of non-gap chars at position A[j-1] (except f_1A[0], which = k ???)
  f_1B      = aa_malloc((n+1) * sizeof(int)); //f_1B[j] = number of non-gap chars at position B[j-1] (except f_1B[0], which = 1 ???)
  f_10A     = aa_malloc((m+1) * sizeof(int)); //f_10A[j] = number of strings that have gap chars at position A[j-1] but non-gap at A[j-2]
  f_10B     = aa_malloc((n+1) * sizeof(int)); //f_10B[j] = number of strings that have gap chars at position B[j-1] but non-gap at B[j-2]
  f_01A     = aa_malloc((m+1) * sizeof(int)); //gap at j-2, non-gap at j-1
  f_01B     = aa_malloc((n+1) * sizeof(int)); //gap at j-2, non-gap at j-1
  f_00A     = aa_malloc((m+1) * sizeof(int)); //gap at j-2 and j-1
  f_00B     = aa_malloc((n+1) * sizeof(int)); //gap at j-2 and j-1
  f_firstA  = aa_malloc((m+1) * sizeof(int)); //f_firstA[j] = number of strings in A whose 1st non-gap char is at column j-1
  f_firstB  = aa_malloc((n+1) * sizeof(int)); //f_firstB[j] = number of strings in B whose 1st non-gap char is at column j-1
  gapstateH = aa_malloc((k+l) * sizeof(int));
  gapstateV = aa_malloc((k+l) * sizeof(int));
  gapstateD = aa_malloc((k+l) * sizeof(int));

  /* per-column count of gap and non-gap chars */
  f_0A[0] = 0; 
  f_1A[0] = k;
  for (j = 1; j <= m; j++) {
    f_0A[j] = 0;
    f_1A[j] = 0;
    for (i = 0; i < k; i++) {
      if (A[i][j-1] == AA_EPS) { 
      	f_0A[j]++; 
      } else {
      	f_1A[j]++;
      }
    }
  }
  f_0B[0] = 0;
  f_1B[0] = l;
  for (j = 1; j <= n; j++) {
    f_0B[j] = 0;
    f_1B[j] = 0;
    for (i = 0; i < l; i++) {
      if (B[i][j-1] == AA_EPS) { 
      	f_0B[j]++; 
      } else {
      	f_1B[j]++;
      }
    }
  }


  /*per-column count of strings whose first letter appears in that column*/
  f_firstA[0] = 0;
  for (i = 1; i <= m; i++) {
    f_firstA[i] = 0;
  }
  for (i = 0; i < k; i++) {
    flag = AA_TRUE;
    for (j = 0; (j < m) && flag; j++) {
      if (A[i][j] != AA_EPS) { f_firstA[j+1]++; flag = AA_FALSE; }
    }
  }
  f_firstB[0] = 0;
  for (i = 0; i <= n; i++) {
    f_firstB[i] = 0;
  }
  for (i = 0; i < l; i++) {
    flag = AA_TRUE;
    for (j = 0; (j < n) && flag; j++) {
      if (B[i][j] != AA_EPS) { f_firstB[j+1]++; flag = AA_FALSE; }
    }
  }
  
  
  /* number of strings that have gap chars at position *j-1* but non-gap at *j-2* */
  f_10A[0] = 0;  
  f_10A[1] = f_0A[1]; //number of gap chars at position A[0]
  for (j = 2; j <= m; j++) {
    f_10A[j] = 0;
    for (i = 0; i < k; i++) {
      if ((A[i][j-2] != AA_EPS) && (A[i][j-1] == AA_EPS)) { f_10A[j]++; }
    }
  }
  f_10B[0] = 0;
  f_10B[1] = f_0B[1];
  for (j = 2; j <= n; j++) {
    f_10B[j] = 0;
    for (i = 0; i < l; i++) {
      if ((B[i][j-2] != AA_EPS) && (B[i][j-1] == AA_EPS)) { f_10B[j]++; }
    }
  }
  
  
  /*number of strings wth gap at j-2 and either gap (f_00[AB]) or non-gap (f_01[AB]) at j-1 */
  f_01A[0] = 0;
  f_01A[1] = f_1A[1]; //number non-gap chars at pos'n A[0]
  f_00A[0] = f_00A[1] = 0;
  for (j = 2; j <= m; j++) {
    f_01A[j] = 0;
    f_00A[j] = 0;
    for (i = 0; i < k; i++) {
      if ((A[i][j-2] == AA_EPS) && (A[i][j-1] != AA_EPS)) { f_01A[j]++; }  //gap at j-2, non-gap at j-1
      if ((A[i][j-2] == AA_EPS) && (A[i][j-1] == AA_EPS)) { f_00A[j]++; }  //gap at j-2 and j-1
    }
  }
  f_01B[0] = 0;
  f_01B[1] = f_1B[1];
  f_00B[0] = f_00B[1] = 0;
  for (j = 2; j <= n; j++) {
    f_01B[j] = 0;
    f_00B[j] = 0;
    for (i = 0; i < l; i++) {
      if ((B[i][j-2] == AA_EPS) && (B[i][j-1] != AA_EPS)) { f_01B[j]++; }
      if ((B[i][j-2] == AA_EPS) && (B[i][j-1] == AA_EPS)) { f_00B[j]++; }
    }
  }

  /* allocate and fill in boundary entries for dynamic programming table */
  H = aa_malloc((m+1) * sizeof(int *));
  V = aa_malloc((m+1) * sizeof(int *));
  D = aa_malloc((m+1) * sizeof(int *));
  for (i = 0; i <= m; i++) {
    H[i] = aa_malloc((n+1) * sizeof(int));
    V[i] = aa_malloc((n+1) * sizeof(int));
    D[i] = aa_malloc((n+1) * sizeof(int));
  }


  H[0][0] = aa_gamma * f_0A[1] * f_1B[1];
  V[0][0] = aa_gamma * f_1A[1] * f_0B[1];
  D[0][0] = 0;

  H[0][1] = (aa_lambda + aa_gamma) * k * f_1B[1];
  V[0][1] = H[0][1] + aa_gamma  * l * f_firstA[1];
  D[0][1] = V[0][1];

  V[1][0] = (aa_lambda + aa_gamma) * l * f_1A[1];
  H[1][0] = V[1][0] + aa_gamma  * k * f_firstB[1];
  D[1][0] = H[1][0];

  for (j = 2; j <= n; j++) {
    H[0][j] =
      H[0][j-1] + aa_lambda * k * f_1B[j]  + aa_gamma * k * f_firstB[j];
    V[0][j] = H[0][j] + aa_gamma * l * f_firstA[1];
    D[0][j] = V[0][j];
  }
  for (i = 2; i <= m; i++) {
    V[i][0] = 
      V[i-1][0] + aa_lambda * l * f_1A[i]  + aa_gamma * l * f_firstA[i];
    H[i][0] = V[i][0] + aa_gamma * k * f_firstB[1];
    D[i][0] = H[i][0];
  }
  
  // fill in the dynamic programming table (row major order)
  for (i = 1; i <= m; i++) {
    for (j = 1; j <= n; j++) {
      a = H[i][j-1] + r * (aa_gamma * k * f_01B[j]);
      b = V[i][j-1] + aa_gamma * (f_1A[i] * f_1B[j] + r * (f_0A[i]*f_1B[j])) ;
      c = D[i][j-1] + aa_gamma * (f_1A[i] * f_1B[j] + r * (f_0A[i]*f_01B[j]));
      min = (a < b)?a:b;
      if (c < min) { min = c; }
      H[i][j] = aa_lambda * k * f_1B[j] + min;

      a = H[i-1][j] + aa_gamma * (f_1A[i] * f_1B[j] + r * (f_1A[i]*f_0B[j]));
      b = V[i-1][j] + r * (aa_gamma * l * f_01A[i]);
      c = D[i-1][j] + aa_gamma * (f_1A[i] * f_1B[j] + r * (f_01A[i]*f_0B[j]));
      min = (a < b)?a:b;
      if (c < min) { min = c; }
      V[i][j] = aa_lambda * l * f_1A[i] + min;

      a = H[i-1][j-1] + aa_gamma 
				* (f_10B[j] * f_1A[i] + r * (f_00B[j]*f_1A[i])
		 	   	+ r * (f_01B[j] * f_0A[i]));
      b = V[i-1][j-1] + aa_gamma 
				* (f_10A[i] * f_1B[j] + r * (f_00A[i]*f_1B[j])
	   			+ r * (f_01A[i] * f_0B[j]));      
      c = D[i-1][j-1] + aa_gamma
				* (f_1A[i] * f_10B[j] + f_10A[i] * f_1B[j]
	   			+ r * (f_00A[i] * f_01B[j] + f_01A[i] * f_00B[j]));

      min = (a < b)?a:b;
      if (c < min) { min = c; }
      sub = 0;
      for (x = 0; x < k; x++) {
		for (y = 0; y < l; y++) {
			if ( A[x][i-1] != AA_EPS && B[y][j-1] != AA_EPS ) {
				int a_ = (int)(A[x][i-1]);
				int b_ = (int)(B[y][j-1]);
				int aa_ = aa_alphaconv[a_];
				int bb_ = aa_alphaconv[b_]	;			
				int cst = aa_cost[aa_][bb_];
				sub += cst;
//				sub += aa_cost[aa_alphaconv[(int)(A[x][i-1])]][aa_alphaconv[(int)(B[y][j-1])]];
			} else if ( A[x][i-1] != AA_EPS || B[y][j-1] != AA_EPS ) {
				sub += aa_lambda;	
			}
		}
      }
      D[i][j] = sub + min;
    }
  }

  a = H[m][n];
  b = V[m][n];
  c = D[m][n];
  if (a<b) {     min = a; j = 1; } 
  else {         min = b; j = 2; }
  if (c < min) { min = c; j = 3; }
  qc = min;

  /* construct the alignment (to calculate actual aa_cost)            */
  /* first pass to calculate length (for allocation), then recover */
  len = 0; x = m; y = n;
  while (x > 0 && y > 0) {
    len++;
    switch (j) {
    case 1: 
      a = H[x][y-1] + r * (aa_gamma * k * f_01B[y]);
      b = V[x][y-1] + aa_gamma * (f_1A[x] * f_1B[y] + r * (f_0A[x]*f_1B[y]));
      c = D[x][y-1] + aa_gamma * (f_1A[x] * f_1B[y] + r * (f_0A[x] * f_01B[y]));
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      y--;
      break;

    case 2:
      a = H[x-1][y] + aa_gamma * (f_1A[x] * f_1B[y] + r * (f_1A[x]*f_0B[y]));
      b = V[x-1][y] + r * (aa_gamma * l * f_01A[x]);
      c = D[x-1][y] + aa_gamma * (f_1A[x] * f_1B[y] + r * (f_01A[x]*f_0B[y]));
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--;
      break;

    case 3:
      a = H[x-1][y-1] 
			+ aa_gamma * (f_10B[y] * f_1A[x] + r * (f_00B[y]*f_1A[x])
		    + r * (f_01B[y] * f_0A[x]));
      b = V[x-1][y-1] 
			+ aa_gamma * (f_10A[x] * f_1B[y] + r * (f_00A[x]*f_1B[y])
		    + r * (f_01A[x] * f_0B[y]));
      c = D[x-1][y-1] 
			+ aa_gamma * (f_1A[x] * f_10B[y] + f_10A[x] * f_1B[y]
		    + r * (f_00A[x] * f_01B[y] + f_01A[x] * f_00B[y]));
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--; y--;
      break;
    };
  }

  if (0 < x)      { len += x; }
  else if (0 < y) { len += y; }



  al = aa_malloc((k+l) * sizeof(char *));
  for (i = 0; i < k+l; i++) {
    al[i] = aa_malloc((len+1) * sizeof(char));
  }

  if (script != NULL) {
	  *script = aa_malloc( len * sizeof(unsigned char));
  }
  
  a = H[m][n]; b = V[m][n]; c = D[m][n];
  if (a<b) { min = a; j = 1; } else { min = b; j = 2; }
  if (c < min) { min = c; j = 3; }


  sub = len; x = m; y = n;
  while (x > 0 && y > 0) {
    switch (j) {
    case 1: 
      if (script != NULL) (*script)[len-1] = AA_INS;
      for (i = 0; i < k; i++) {	al[i][len-1] = AA_EPS; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = B[i][y-1]; }
      a = H[x][y-1] + r * (aa_gamma * k * f_01B[y]);
      b = V[x][y-1] + aa_gamma * (f_1A[x] * f_1B[y] + r * (f_0A[x]*f_1B[y]));
      c = D[x][y-1] + aa_gamma * (f_1A[x] * f_1B[y] + r * (f_0A[x] * f_01B[y]));
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      y--;
      break;

    case 2:
      if (script != NULL) (*script)[len-1] = AA_DEL;
      for (i = 0; i < k; i++) {	al[i][len-1] = A[i][x-1]; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = AA_EPS; }
      a = H[x-1][y] + aa_gamma * (f_1A[x] * f_1B[y] + r * (f_1A[x]*f_0B[y]));
      b = V[x-1][y] + r * (aa_gamma * l * f_01A[x]);
      c = D[x-1][y] + aa_gamma * (f_1A[x] * f_1B[y] + r * (f_01A[x]*f_0B[y]));
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--;
      break;

    case 3:
      if (script != NULL) (*script)[len-1] = AA_SUB;
      for (i = 0; i < k; i++) {	al[i][len-1] = A[i][x-1]; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = B[i][y-1]; }
      a = H[x-1][y-1] 
      		+ aa_gamma * (f_10B[y] * f_1A[x] + r * (f_00B[y]*f_1A[x])
		    + r * (f_01B[y] * f_0A[x]));
      b = V[x-1][y-1] 
			+ aa_gamma * (f_10A[x] * f_1B[y] + r * (f_00A[x]*f_1B[y])
		    + r * (f_01A[x] * f_0B[y]));
      c = D[x-1][y-1]
			+ aa_gamma * (f_1A[x] * f_10B[y] + f_10A[x] * f_1B[y]
		    + r * (f_00A[x] * f_01B[y] + f_01A[x] * f_00B[y]));
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--; y--;
      break;
    };
    len--;
  }

  if (0 < x)      { 
    while (0 < x) {
      if (script != NULL) (*script)[len-1] = AA_DEL;
      for (i = 0; i < k; i++) {	al[i][len-1] = A[i][x-1]; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = AA_EPS; }
      x--;
      len--;
    }
  }
  else if (0 < y) {
    while (0 < y) {
      if (script != NULL) (*script)[len-1] = AA_INS;
      for (i = 0; i < k; i++) {	al[i][len-1] = AA_EPS; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = B[i][y-1]; }
      y--;
      len--;
    }
  }

  len = sub;
  c = alignmentCost(al, al + k, k, l, len, len);

  *return_alignment = al;
  *return_len = len;

  aa_free(f_0A, (m+1) * sizeof(int));
  aa_free(f_0B, (m+1) * sizeof(int));
  aa_free(f_1A, (m+1) * sizeof(int));
  aa_free(f_1B, (m+1) * sizeof(int));
  aa_free(f_10A, (m+1) * sizeof(int));
  aa_free(f_10B, (m+1) * sizeof(int));
  aa_free(f_01A, (m+1) * sizeof(int));
  aa_free(f_01B, (m+1) * sizeof(int));
  aa_free(f_00A, (m+1) * sizeof(int));
  aa_free(f_00B, (m+1) * sizeof(int));
  aa_free(f_firstA, (m+1) * sizeof(int));
  aa_free(f_firstB, (m+1) * sizeof(int));
  for (i = 0; i <= m; i++) {
    aa_free(H[i], (n+1) * sizeof(int));
    aa_free(V[i], (n+1) * sizeof(int));
    aa_free(D[i], (n+1) * sizeof(int));
  }
  aa_free(H, (m+1) * sizeof(int *));
  aa_free(V, (m+1) * sizeof(int *));
  aa_free(D, (m+1) * sizeof(int *));

  aa_free(gapstateH, (k+l) * sizeof(int));
  aa_free(gapstateV, (k+l) * sizeof(int));
  aa_free(gapstateD, (k+l) * sizeof(int));
  
  return c;

}

/**********************************************************************
 * procedure: pessimistic_optimistic_weighted
 *   Input:
 *     aa_A, aa_B: multiple sequence alignments
 *     k, l: number of rows in aa_A and aa_B respectively
 *     m, n: number of columns in aa_A and aa_B respectively
 *        d: substitution aa_cost matrix
 *        r: real value in the range [0,1], r*aa_gamma is charged for a 
 *           gap in the case where itcan't be determined if a gap opens.
 *   Output: qc, a int representing the actual aa_cost of the optimal
 *           hueristic *weighted* alignment of aa_A and aa_B using pessimistic or optimistc
 *           (or between) gap counts (optimistic if r = 0, pessimistic if
 *           r = 1, and whatever you want to call it if 0 < r < 1.
 *           
 *********************************************************************/
int
pessimistic_optimistic_incl_alignment_weighted (char **A, char **B, 
			int k, int l, int m, int n, double r, char ***return_alignment, 
			int *return_len, unsigned char **script ) {

  int     i, j, x, y, p, q, len;
  int   a, b, c, d, sub;
  int    *firstA, *firstB, *gapstateH,*gapstateV, *gapstateD;
  int **H, **V, **D;
  char  **al;
  int  min, qc;

  firstA  = aa_malloc((k) * sizeof(int)); //firstA[j] = the column in which the first letter appears for string A[j]
  firstB  = aa_malloc((l) * sizeof(int)); //firstB[j] = the column in which the first letter appears for string B[j]
  gapstateH = aa_malloc((k+l) * sizeof(int));
  gapstateV = aa_malloc((k+l) * sizeof(int));
  gapstateD = aa_malloc((k+l) * sizeof(int));


  /* allocate and fill in boundary entries for dynamic programming table */
  H = aa_malloc((m+1) * sizeof(int *));
  V = aa_malloc((m+1) * sizeof(int *));
  D = aa_malloc((m+1) * sizeof(int *));
  for (i = 0; i <= m; i++) {
    H[i] = aa_malloc((n+1) * sizeof(int));
    V[i] = aa_malloc((n+1) * sizeof(int));
    D[i] = aa_malloc((n+1) * sizeof(int));
  }

  for (i=0; i<=m; i++) {
  	for (j=0; j<=n; j++) {
	  H[i][j] = V[i][j] = D[i][j] = 0;
  	}
  }

  //fill in H[0][0] and V[0][0]
  for (p=0; p<k; p++) {
  	for (q=0; q<l; q++) {
  		if (A[p][0] == AA_EPS  && B[q][0] != AA_EPS) {
			H[0][0] +=  aa_weights[p][k+q] * GET_GAMMA_A(p,0) ;
  		} else if (A[p][0] != AA_EPS  && B[q][0] == AA_EPS) {
	  		V[0][0] +=  aa_weights[p][k+q] * GET_GAMMA_B(q,0);  
  		}
  	}
  }



  //fill in H[0][1], V[0][1] & D[0][1]
  for (p=0; p<k; p++) {
	 for (q=0; q<l; q++) {
  		if (B[q][0] != AA_EPS) {
			H[0][1] += aa_weights[p][k+q] * (aa_lambda_terminal + aa_gamma_terminal ) ; // gap in A appears before the first column of A - must be terminal
  		}
  		if (aa_firstCharsA[p] == 0) {//if posn 0 is the first non-gap in string A[i]) 
  			                 //- need this because an extension out of this cell in the vertical direction (to 1,1) won't incur the gap open at that time, so we incur it here
  			V[0][1] += aa_weights[p][k+q] * GET_GAMMA_B(q,0); // if B[0] was a letter, new gap of B against A will be internal
  		}
	 }
  }
  V[0][1] += H[0][1];
  D[0][1] = V[0][1];
  
  //fill in H[1][0], V[1][0] & D[1][0]
  for (p=0; p<k; p++) {
	 for (q=0; q<l; q++) {
  		if (A[p][0] != AA_EPS) {
			V[1][0] += aa_weights[p][k+q] * ( aa_lambda_terminal + aa_gamma_terminal ) ; // gap in B appears before the first column of B - must be terminal
  		}
  		if (aa_firstCharsB[q] == 0) { //posn 0 is the first non-gap in string B[j]) {
   			                 //- need this because an extension out of this cell in the horizontal direction (to 1,1) won't incur the gap open at that time, so we incur it here
  			H[1][0] += aa_weights[p][k+q] * GET_GAMMA_A(p,0); // if A[0] was a letter, new gap of A against B will be internal
  		}
	 }
  }
  H[1][0] += V[1][0];
  D[1][0] = H[1][0];
	  	  
  for (i = 2; i <= n; i++) {
  	H[0][i] = H[0][i-1];
	for (p=0; p<k; p++) {
		for (q=0; q<l; q++) {
	  		if (B[q][i-1] != AA_EPS) {
		    	H[0][i] += aa_weights[p][k+q] * aa_lambda_terminal; // gap in A appears before the first column of A - must be terminal
	  		}
			if (aa_firstCharsB[q] == i-1) { //posn x-1 is the first non-gap in string B[j]) {
				H[0][i] += aa_weights[p][k+q] * aa_gamma_terminal; // gap in A appears before the first column of A - must be terminal
			}
			if (aa_firstCharsA[p] == 0) { //posn 0 is the first non-gap in string A[j]) {
				V[0][i] += aa_weights[p][k+q] * GET_GAMMA_B(q,i-1);
			}
		}
	}
    V[0][i] += H[0][i];
    D[0][i] = V[0][i];    
  }
  for (i = 2; i <= m; i++) {
  	V[i][0] = V[i-1][0];
	for (p=0; p<k; p++) {
		for (q=0; q<l; q++) {
	  		if (A[p][i-1] != AA_EPS) {
		    	V[i][0] += aa_weights[p][k+q] * aa_lambda_terminal ;  // gap in B appears before the first column of B - must be terminal
	  		}
			if (aa_firstCharsA[p]  == i-1) { //posn x-1 is the first non-gap in string A[i]) {
				V[i][0] += aa_weights[p][k+q] * aa_gamma_terminal ;  // gap in A appears before the first column of A - must be terminal
			}
			if (aa_firstCharsB[q] /*firstB[q]*/ == 0) { //posn 0 is the first non-gap in string B[j]) {
				H[i][0] += aa_weights[p][k+q] * GET_GAMMA_A(p,i-1);
			}
		}
	}
    H[i][0] += V[i][0];
    D[i][0] = H[i][0];    
  }

	
  // fill in the dynamic programming table (row major order)
  for (i = 1; i <= m; i++) {
    for (j = 1; j <= n; j++) {
		calcCost ( H, V, D, A, B, AA_HORIZ, i, j, k, l, r, &a, &b, &c, &d);
		H[i][j] = d + minABC(a,b,c);
      	
      	calcCost ( H, V, D, A, B, AA_VERT, i, j, k, l, r, &a, &b, &c, &d);
		V[i][j] = d + minABC(a,b,c);

		calcCost ( H, V, D, A, B, AA_DIAG, i, j, k, l, r, &a, &b, &c, &d);
		D[i][j] = d + minABC(a,b,c);
    }
  }

  a = H[m][n];
  b = V[m][n];
  c = D[m][n];
  if (a<b) {     min = a; j = 1; } 
  else {         min = b; j = 2; }
  if (c < min) { min = c; j = 3; }
  qc = min;

  /* construct the alignment (to calculate actual aa_cost)            */
  /* first pass to calculate length (for allocation), then recover */
  len = 0; x = m; y = n;
  while (x > 0 && y > 0) {
    len++;
    switch (j) {
    case 1: 
 	  calcCost ( H, V, D, A, B, AA_HORIZ, x, y, k, l, r, &a, &b, &c, &d);
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      y--;
      break;

    case 2:
 	  calcCost ( H, V, D, A, B, AA_VERT, x, y, k, l, r, &a, &b, &c, &d);    
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--;
      break;

    case 3:
 	  calcCost ( H, V, D, A, B, AA_DIAG, x, y, k, l, r, &a, &b, &c, &d);
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--; y--;
      break;
    };
  }

  if (0 < x)      { len += x; }
  else if (0 < y) { len += y; }



  al = aa_malloc((k+l) * sizeof(char *));
  for (i = 0; i < k+l; i++) {
    al[i] = aa_malloc((len+1) * sizeof(char));
  }

  if (script != NULL) {
	  *script = aa_malloc( len * sizeof(unsigned char));
  }
  
  a = H[m][n]; b = V[m][n]; c = D[m][n];
  if (a<b) { min = a; j = 1; } else { min = b; j = 2; }
  if (c < min) { min = c; j = 3; }

  sub = len; x = m; y = n;
  while (x > 0 && y > 0) {
    switch (j) {
    case 1: 
      if (script != NULL) (*script)[len-1] = AA_INS;
      for (i = 0; i < k; i++) {	al[i][len-1] = AA_EPS; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = B[i][y-1]; }
 	  calcCost ( H, V, D, A, B, AA_HORIZ, x, y, k, l, r, &a, &b, &c, &d);
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      y--;
      break;

    case 2:
      if (script != NULL) (*script)[len-1] = AA_DEL;
      for (i = 0; i < k; i++) {	al[i][len-1] = A[i][x-1]; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = AA_EPS; }
 	  calcCost ( H, V, D, A, B, AA_VERT, x, y, k, l, r, &a, &b, &c, &d);
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--;
      break;

    case 3:
      if (script != NULL) (*script)[len-1] = AA_SUB;
      for (i = 0; i < k; i++) {	al[i][len-1] = A[i][x-1]; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = B[i][y-1]; }
 	  calcCost ( H, V, D, A, B, AA_DIAG, x, y, k, l, r, &a, &b, &c, &d);
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--; y--;
      break;
    };
    len--;
  }

  if (0 < x)      { 
    while (0 < x) {
      if (script != NULL) (*script)[len-1] = AA_DEL;
      for (i = 0; i < k; i++) {	al[i][len-1] = A[i][x-1]; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = AA_EPS; }
      x--;
      len--;
    }
  }
  else if (0 < y) {
    while (0 < y) {
      if (script != NULL) (*script)[len-1] = AA_INS;
      for (i = 0; i < k; i++) {	al[i][len-1] = AA_EPS; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = B[i][y-1]; }
      y--;
      len--;
    }
  }

  len = sub;
  c = alignmentCost(al, al + k, k, l, len, len);

  *return_alignment = al;
  *return_len = len;

  aa_free(firstA, (k) * sizeof(int));
  aa_free(firstB, (l) * sizeof(int));
  for (i = 0; i <= m; i++) {
    aa_free(H[i], (n+1) * sizeof(int));
    aa_free(V[i], (n+1) * sizeof(int));
    aa_free(D[i], (n+1) * sizeof(int));
  }
  aa_free(H, (m+1) * sizeof(int *));
  aa_free(V, (m+1) * sizeof(int *));
  aa_free(D, (m+1) * sizeof(int *));

  aa_free(gapstateH, (k+l) * sizeof(int));
  aa_free(gapstateV, (k+l) * sizeof(int));
  aa_free(gapstateD, (k+l) * sizeof(int));
  
  return c;
}


void
calcCost ( int **H, int **V, int **D, 
			char **A, char **B, int endDirec, 
			int i, int j, int k, int l, double r,
			int *a, int *b, int *c, int *d ){
    int p, q;
    *d = 0;
	if (endDirec == AA_HORIZ) {
    	*a = H[i][j-1];
     	*b = V[i][j-1];
	    *c = D[i][j-1];
	} else if (endDirec == AA_VERT) {
		*a = H[i-1][j];
		*b = V[i-1][j];
		*c = D[i-1][j];
	} else if (endDirec == AA_DIAG) {
		*a = H[i-1][j-1];
		*b = V[i-1][j-1];
		*c = D[i-1][j-1];
	}
	
	for (p=0; p<k; p++) {
		for (q=0; q<l; q++) {
			if (endDirec == AA_HORIZ) {
				if ( B[q][j-1] != AA_EPS ) {
					int loc_gamma = GET_GAMMA_A(p,i-1);
					//add to a  (from horizontal)
			  		if ( j==1 || B[q][j-2] == AA_EPS ) { //Note: This could double count gap opens (possibly loosening the bound this function returns):
			  											//   the j==1 case is the 2nd column of the DP table ... and the first column (H[0][j]) already has the gap start costs factored in
				    	*a += r * aa_weights[p][k+q] * loc_gamma;
				    	//TO DO: could actually tighten the pessimistic bound a bit? - no need to count duplicate gap open costs if we haven't even reached i>lastCharA
			  		}
					//add to b (from vertical)
		 			if ( A[p][i-1] != AA_EPS ) {
				    	*b += aa_weights[p][k+q] * loc_gamma; //i-1 might be the final letter of the string
		 			} else {
				    	*b += r * aa_weights[p][k+q] * loc_gamma;
		 			}
					//add to c  (from diagonal)
					if ( A[p][i-1] != AA_EPS ) {
				    	*c += aa_weights[p][k+q] * loc_gamma;
					} else if ( j==1 || B[q][j-2] == AA_EPS ) { // the j==1 case is the 2nd column of the DP table ... and the first column (D[0][j]) already has the gap start costs factored in
				    	*c += r * aa_weights[p][k+q] * loc_gamma;
		 			}
					//add to H[i][j]				
					*d += aa_weights[p][k+q] * GET_LAMBDA_A(p,i-1) ;
				}
			} else if (endDirec == AA_VERT) {
				if ( A[p][i-1] != AA_EPS ) {
					int loc_gamma = GET_GAMMA_B(q,j-1);
					//add to a
		 			if ( B[q][j-1] != AA_EPS ) {
				    	*a += aa_weights[p][k+q] * loc_gamma;
		 			} else {
				    	*a += r * aa_weights[p][k+q] * loc_gamma;     				
		 			}
					//add to b
			  		if ( i==1 || A[p][i-2] == AA_EPS ) { 
				    	*b += r * aa_weights[p][k+q] * loc_gamma;
			  		}
					//add to c				
					if ( B[q][j-1] != AA_EPS ) {
				    	*c += aa_weights[p][k+q] * loc_gamma;
					} else if (  (i==1 || A[p][i-2] == AA_EPS)  &&  (B[q][j-1] == AA_EPS) ) {
				    	*c += r * aa_weights[p][k+q] * loc_gamma;     				
		 			}
					//add to V[i][j]				
					*d += aa_weights[p][k+q] * GET_LAMBDA_B(q,j-1) ;
				}
			} else if (endDirec == AA_DIAG) {
				int loc_gamma_a = GET_GAMMA_A(p,i-1);
				int loc_gamma_b = GET_GAMMA_B(q,j-1);

				//add to a	 (from horizontal)
				if ( A[p][i-1] != AA_EPS  && B[q][j-1] == AA_EPS) {
		 			if ( j==1  || B[q][j-2] != AA_EPS ) {
				    	*a += aa_weights[p][k+q] * loc_gamma_b;
		 			} else {
				    	*a += r * aa_weights[p][k+q] * loc_gamma_b;     				
		 			}
				} else if ( A[p][i-1] == AA_EPS && B[q][j-1] != AA_EPS && (j==1 || B[q][j-2] == AA_EPS) ) {
				   	*a += r * aa_weights[p][k+q] * loc_gamma_a;
				}
				//add to b (from vertical)
				if ( A[p][i-1] == AA_EPS && B[q][j-1] != AA_EPS ) {
			  		if ( i==1 || A[p][i-2] != AA_EPS) {
		  				*b += aa_weights[p][k+q] * loc_gamma_a;
			  		} else {
				    	*b += r * aa_weights[p][k+q] * loc_gamma_a;
			  		}
				} else if (  A[p][i-1] != AA_EPS && B[q][j-1] == AA_EPS && (i==1 || A[p][i-2]  == AA_EPS) ) {
			    	*b += r * aa_weights[p][k+q] * loc_gamma_b;					
				}
				//add to c	(from diagonal)			
				if ( A[p][i-1] != AA_EPS && B[q][j-1] == AA_EPS ) {	
					if ( j==1 || B[q][j-2] != AA_EPS ) {
				    	*c += aa_weights[p][k+q] * loc_gamma_b;
					} else if ( i==1 || A[p][i-2] == AA_EPS ) {
				    	*c += r * aa_weights[p][k+q] * loc_gamma_b;						
					}	
				} else if ( A[p][i-1] == AA_EPS && B[q][j-1] != AA_EPS ) {	
					if ( i==1 || A[p][i-2] != AA_EPS ) {
				    	*c += aa_weights[p][k+q] * loc_gamma_a;
					} else if ( j==1 || B[q][j-2] == AA_EPS ) {
				    	*c += r * aa_weights[p][k+q] * loc_gamma_a;			
					}					    	
				}
				//add to D[i][j]
				if ( A[p][i-1] != AA_EPS && B[q][j-1] != AA_EPS ) {
					*d += aa_weights[p][k+q] * aa_cost[aa_alphaconv[(int)(A[p][i-1])]][aa_alphaconv[(int)(B[q][j-1])]];
				} else if ( A[p][i-1] != AA_EPS ) {
					*d += aa_weights[p][k+q] * GET_LAMBDA_B(q,j-1);
				} else if ( B[q][j-1] != AA_EPS ) {
					*d += aa_weights[p][k+q] * GET_LAMBDA_A(p,i-1);
				}
		    }
		}
	}
}

int
pessimistic_optimistic (char **A, char **B, 
			int k, int l, int m, int n, double r) {

	int i, len;
	char **alignment;
	int cost = pessimistic_optimistic_incl_alignment (A, B, k, l, m, n, r, &alignment, &len, NULL);
	
	for (i = 0; i < k+l; i++) { aa_free(alignment[i], (len+1) * sizeof(char)); }
	aa_free(alignment, (k+l) * sizeof(char));
	
	return cost;
}


/**********************************************************************
 * procedure: reverse_optimistic_table
 *   Input:
 *     aa_A, aa_B: multiple sequence alignments
 *     m, n: lengths of aa_A and aa_B respectively
 *     k, l: numbers of sequences in aa_A and aa_B respectively
 *        d: substitution aa_cost matrix
 *   Output: oc, an integer representing the aa_cost of the optimal 
 *           alignment of aa_A and aa_B using optimistic gap counts.
 *           H, V, D, the dynamic programming table
 *********************************************************************/
int
reverse_optimistic_table (char **A, char **B, int k, int l, int m, int n, 
		  int ***H, int ***V, int ***D) {

	int x; 
	if (aa_weighted || aa_gamma != aa_gamma_terminal  || aa_lambda != aa_lambda_terminal ) {
		x = reverse_optimistic_table_weighted (A, B, k, l, m, n, H, V, D);
	} else {
		x = reverse_optimistic_table_unweighted (A, B, k, l, m, n, H, V, D);		
	}

/*	int i,j;
	printf("REVERSE TABLES\n");
	printf ("H\n");
	for (i=0; i<=m; i++) {
		for (j=0; j<=n; j++) {		
			printf ("%d\t", (*H)[i][j]);		
		}
		printf("\n");
	}
	printf("\n");
	printf ("V\n");
	for (i=0; i<=m; i++) {
		for (j=0; j<=n; j++) {		
			printf ("%d\t", (*V)[i][j]);		
		}
		printf("\n");
	}
	printf("\n");	
	printf ("D\n");
	for (i=0; i<=m; i++) {
		for (j=0; j<=n; j++) {		
			printf ("%d\t", (*D)[i][j]);		
		}
		printf("\n");
	}	
	printf("\n");
	fflush (stdout);
*/	
	return x;
}


/**********************************************************************
 * procedure: reverse_optimistic_table_unweighted
 *   Input:
 *     aa_A, aa_B: multiple sequence alignments
 *     m, n: lengths of aa_A and aa_B respectively
 *     k, l: numbers of sequences in aa_A and aa_B respectively
 *        d: substitution aa_cost matrix
 *   Output: oc, an integer representing the aa_cost of the optimal 
 *           alignment of aa_A and aa_B using optimistic gap counts.
 *           H, V, D, the dynamic programming table
 *********************************************************************/
int
reverse_optimistic_table_unweighted (char **A, char **B, int k, int l, int m, int n, 
		  int ***H, int ***V, int ***D) {
  
  int     i, j, x, y, flag;
  int   a, b, c, oc, min, sub;
  int    *f_1A, *f_1B, *f_firstA, *f_firstB, *f_10A, *f_10B;
  int     len;
  char  **al;
  char  **Arev, **Brev;
  int **HR, **VR, **DR; 

  Arev = aa_malloc(k * sizeof(char *));
  for (i = 0; i < k; i++) { Arev[i] = aa_malloc((m+1) * sizeof(char)); }
  Brev = aa_malloc(l * sizeof(char *));
  for (i = 0; i < l; i++) { Brev[i] = aa_malloc((n+1) * sizeof(char)); }

  for (i = 0; i < k; i++) {
    for (j = 0; j < m; j++) {
      Arev[i][j] = A[i][m-j-1];
    }
    Arev[i][m] = '\0';
  }
  for (i = 0; i < l; i++) {
    for (j = 0; j < n; j++) {
      Brev[i][j] = B[i][n-j-1];
    }
    Brev[i][n] = '\0';
  }
 
  f_1A = aa_malloc((m+1) * sizeof(int));
  f_1B = aa_malloc((n+1) * sizeof(int));
  f_firstA = aa_malloc((m+1) * sizeof(int));
  f_firstB = aa_malloc((n+1) * sizeof(int));
  f_10A = aa_malloc((m+1) * sizeof(int));
  f_10B = aa_malloc((n+1) * sizeof(int));

  f_1A[0] = k;
  for (j = 1; j <= m; j++) {
    f_1A[j] = 0;
    for (i = 0; i < k; i++) {
      if (Arev[i][j-1] != AA_EPS) { f_1A[j]++; }
    }
  }
  f_1B[0] = l;
  for (j = 1; j <= n; j++) {
    f_1B[j] = 0;
    for (i = 0; i < l; i++) {
      if (Brev[i][j-1] != AA_EPS) { f_1B[j]++; }
    }
  }

  f_firstA[0] = 0;
  for (i = 1; i <= m; i++) {
    f_firstA[i] = 0;
  }
  for (i = 0; i < k; i++) {
    flag = AA_TRUE;
    for (j = 1; (j <= m) && flag; j++) {
      if (Arev[i][j-1] != AA_EPS) { f_firstA[j]++; flag = AA_FALSE; }
    }
  }
  f_firstB[0] = 0;
  for (i = 0; i <= n; i++) {
    f_firstB[i] = 0;
  }
  for (i = 0; i < l; i++) {
    flag = AA_TRUE;
    for (j = 1; (j <= n) && flag; j++) {
      if (Brev[i][j-1] != AA_EPS) { f_firstB[j]++; flag = AA_FALSE; }
    }
  }
  
  f_10A[0] = f_10A[1] = 0;
  for (i = 0; i < k; i++) {
    if (Arev[i][0] == AA_EPS) { f_10A[1]++; }
  }
  for (j = 2; j <= m; j++) {
    f_10A[j] = 0;
    for (i = 0; i < k; i++) {
      if ((Arev[i][j-2] != AA_EPS) && (Arev[i][j-1] == AA_EPS)) { f_10A[j]++; }
    }
  }

  f_10B[0] = f_10B[1] = 0;
  for (i = 0; i < l; i++) {
    if (Brev[i][0] == AA_EPS) { f_10B[1]++; }
  }
  for (j = 2; j <= n; j++) {
    f_10B[j] = 0;
    for (i = 0; i < l; i++) {
      if ((Brev[i][j-2] != AA_EPS) && (Brev[i][j-1] == AA_EPS)) { f_10B[j]++; }
    }
  }

  /* allocate and fill in boundary entries for dynamic programming table */
  *H = aa_malloc((m+1) * sizeof(int *));
  *V = aa_malloc((m+1) * sizeof(int *));
  *D = aa_malloc((m+1) * sizeof(int *));
  for (i = 0; i <= m; i++) {
    (*H)[i] = aa_malloc((n+1) * sizeof(int));
    (*V)[i] = aa_malloc((n+1) * sizeof(int));
    (*D)[i] = aa_malloc((n+1) * sizeof(int));
  }

  HR = aa_malloc((m+1) * sizeof(int *));
  VR = aa_malloc((m+1) * sizeof(int *));
  DR = aa_malloc((m+1) * sizeof(int *));
  for (i = 0; i <= m; i++) {
    HR[i] = aa_malloc((n+1) * sizeof(int));
    VR[i] = aa_malloc((n+1) * sizeof(int));
    DR[i] = aa_malloc((n+1) * sizeof(int));
  }

  // initialize the table
  HR[0][0] = aa_gamma * (k - f_1A[1]) * f_1B[1];
  VR[0][0] = aa_gamma * f_1A[1] * (l - f_1B[1]);
  DR[0][0] = 0;
  
//  printf ("unweighted: HR[0][0] = %d,   VR[0][0] = %d\n", HR[0][0],VR[0][0]);  

  HR[0][1] = (aa_lambda + aa_gamma) * k * f_1B[1];
  VR[0][1] = HR[0][1] + (aa_gamma * l * f_firstA[1]);
  DR[0][1] = VR[0][1];

//  printf ("unweighted: HR[0][1] = %d,   VR[0][1] = %d,   DR[0][1] = %d\n", HR[0][1], VR[0][1], DR[0][1]);

  VR[1][0] = (aa_lambda + aa_gamma) * l * f_1A[1];
  HR[1][0] = VR[1][0] + (aa_gamma * k * f_firstB[1]);
  DR[1][0] = HR[1][0];
//  printf ("unweighted: HR[1][0] = %d,   VR[1][0] = %d,   DR[1][0] = %d\n", HR[1][0], VR[1][0], DR[1][0]);

  for (j = 2; j <= n; j++) {
    HR[0][j] = HR[0][j-1] 
      + aa_lambda * k * f_1B[j] + aa_gamma * k * f_firstB[j];
    VR[0][j] = HR[0][j] + aa_gamma * l * f_firstA[1];
    DR[0][j] = VR[0][j];
  }

  for (i = 2; i <= m; i++) {
    VR[i][0] = VR[i-1][0] 
      + aa_lambda * l * f_1A[i] + aa_gamma * l * f_firstA[i];
    HR[i][0] = VR[i][0] + aa_gamma * k * f_firstB[1];
    DR[i][0] = HR[i][0];
  }

  /* fill in the dynamic programming table (row major order) */
  for (i = 1; i <= m; i++) {
    for (j = 1; j <= n; j++) {
      a = HR[i][j-1];
      b = VR[i][j-1] + aa_gamma * f_1A[i] * f_1B[j];
      c = DR[i][j-1] + aa_gamma * f_1A[i] * f_1B[j];
      if (a < b) { min = a; }
      else { min = b; }
      if (c < min) { min = c; }
      HR[i][j] = aa_lambda * k * f_1B[j] + min;

      a = HR[i-1][j] + aa_gamma * f_1A[i] * f_1B[j];
      b = VR[i-1][j];
      c = DR[i-1][j] + aa_gamma * f_1A[i] * f_1B[j];
      if (a < b) { min = a; }
      else { min = b; }
      if (c < min) { min = c; }
      VR[i][j] = aa_lambda * l * f_1A[i] + min;

      a = HR[i-1][j-1] + aa_gamma * f_1A[i]  * f_10B[j];
      b = VR[i-1][j-1] + aa_gamma * f_10A[i] * f_1B[j];
      c = DR[i-1][j-1] + aa_gamma * (f_1A[i] * f_10B[j] + f_10A[i] * f_1B[j]);
      if (a < b) { min = a; }
      else { min = b; }
      if (c < min) { min = c; }
      sub = 0;
      /* @@@ change to check for alpha size optimization */
      for (x = 0; x < k; x++) {
		for (y = 0; y < l; y++) {
			if ( Arev[x][i-1] != AA_EPS && Brev[y][j-1] != AA_EPS ) {
				sub += aa_cost[aa_alphaconv[(int)Arev[x][i-1]]][aa_alphaconv[(int)Brev[y][j-1]]];
			} else if ( Arev[x][i-1] != AA_EPS || Brev[y][j-1] != AA_EPS ) {
				sub += aa_lambda;	
			}
		}
      }
      DR[i][j] = sub + min;
    }
  }

  a = HR[m][n];
  b = VR[m][n];
  c = DR[m][n];
  if (a < b) {   min = a; j = 1; }
  else {         min = b; j = 2; }
  if (c < min) { min = c; j = 3; }
  oc = min;

  /* construct the alignment (to calculate actual aa_cost)            */
  /* first pass to calculate length (for allocation), then recover */
  /* bug here for aa_gamma = 0, x can become negative for example */
  len = 0; x = m; y = n;
  while (x > 0 && y > 0) {
    len++;
    switch (j) {
    case 1: 
      a = HR[x][y-1];
      b = VR[x][y-1] + aa_gamma * f_1A[x] * f_1B[y];
      c = DR[x][y-1] + aa_gamma * f_1A[x] * f_1B[y];
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      y--;
      break;

    case 2:
      a = HR[x-1][y] + aa_gamma * f_1A[x] * f_1B[y];
      b = VR[x-1][y];
      c = DR[x-1][y] + aa_gamma * f_1A[x] * f_1B[y];
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--;
      break;

    case 3:
      a = HR[x-1][y-1] + aa_gamma * f_1A[x]  * f_10B[y];
      b = VR[x-1][y-1] + aa_gamma * f_10A[x] * f_1B[y];
      c = DR[x-1][y-1] + aa_gamma * (f_1A[x] * f_10B[y] + f_10A[x] * f_1B[y]);
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--; y--;
      break;
    };
  }

  if (0 < x)      { len += x; }
  else if (0 < y) { len += y; }

  /* @@@ make routine to compute actual aa_cost without the actual alignment */
  al = aa_malloc((k+l) * sizeof(char *));
  for (i = 0; i < k+l; i++) { al[i] = aa_malloc((len+1) * sizeof(char)); }
  a = HR[m][n];
  b = VR[m][n];
  c = DR[m][n];
  if (a < b) { min = a; j = 1; } else { min = b; j = 2; }
  if (c < min) { min = c; j = 3; }
  sub = len; x = m; y = n;

  while (x > 0 && y > 0) {
    switch (j) {
    case 1: 
      for (i = 0; i < k; i++) {	al[i][len-1] = AA_EPS; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = Brev[i][y-1]; }
      a = HR[x][y-1];
      b = VR[x][y-1] + aa_gamma * f_1A[x] * f_1B[y];
      c = DR[x][y-1] + aa_gamma * f_1A[x] * f_1B[y];
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      y--;
      break;

    case 2:
      for (i = 0; i < k; i++) {	al[i][len-1] = Arev[i][x-1]; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = AA_EPS; }
      a = HR[x-1][y] + aa_gamma * f_1A[x] * f_1B[y];
      b = VR[x-1][y];
      c = DR[x-1][y] + aa_gamma * f_1A[x] * f_1B[y];
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--;
      break;

    case 3:
      for (i = 0; i < k; i++) {	al[i][len-1] = Arev[i][x-1]; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = Brev[i][y-1]; }
      a = HR[x-1][y-1] + aa_gamma * f_1A[x]  * f_10B[y];
      b = VR[x-1][y-1] + aa_gamma * f_10A[x] * f_1B[y];
      c = DR[x-1][y-1] + aa_gamma * (f_1A[x] * f_10B[y] + f_10A[x] * f_1B[y]);
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--; y--;
      break;
    };
    len--;
  }

  if (0 < x)      { 
    while (0 < x) {
      for (i = 0; i < k; i++) {	al[i][len-1] = Arev[i][x-1]; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = AA_EPS; }
      x--;
      len--;
    }
  }
  else if (0 < y) {
    while (0 < y) {
      for (i = 0; i < k; i++) {	al[i][len-1] = AA_EPS; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = Brev[i][y-1]; }
      y--;
      len--;
    }
  }

  len = sub;
  c = alignmentCost(al, al + k, k, l, len, len);

  for (i = 0; i <= m; i++) {
    for (j = 0; j <=n; j++) {
      (*H)[i][j] = HR[m-i][n-j];
      (*V)[i][j] = VR[m-i][n-j];
      (*D)[i][j] = DR[m-i][n-j];      
    }
  }

  for (i = 0; i < k+l; i++) { aa_free(al[i], (len+1) * sizeof(char)); }
  aa_free(al, (k+l) * sizeof(char));
  aa_free(f_1A, (m+1) * sizeof(int));
  aa_free(f_1B, (m+1) * sizeof(int));
  aa_free(f_10A, (m+1) * sizeof(int));
  aa_free(f_10B, (m+1) * sizeof(int));
  aa_free(f_firstA, (m+1) * sizeof(int));
  aa_free(f_firstB, (m+1) * sizeof(int));
  for (i = 0; i < k; i++) { aa_free(Arev[i], (m+1) * sizeof(char)); }
  aa_free(Arev, k * sizeof(char *));
  for (i = 0; i < l; i++) { aa_free(Brev[i], (n+1) * sizeof(char)); }
  aa_free(Brev, l * sizeof(char *));
  for (i = 0; i <= m; i++) {
    aa_free(HR[i], (n+1) * sizeof(int));
    aa_free(VR[i], (n+1) * sizeof(int));
    aa_free(DR[i], (n+1) * sizeof(int));
  }
  aa_free(HR, (m+1) * sizeof(int *));
  aa_free(VR, (m+1) * sizeof(int *));
  aa_free(DR, (m+1) * sizeof(int *));

  return c;
}

/**********************************************************************
 * procedure: reverse_optimistic_table_weighted
 *   Input:
 *     aa_A, aa_B: multiple sequence alignments
 *     m, n: lengths of aa_A and aa_B respectively
 *     k, l: numbers of sequences in aa_A and aa_B respectively
 *        d: substitution aa_cost matrix
 *   Output: oc, an integer representing the aa_cost of the optimal 
 *           alignment of aa_A and aa_B using optimistic gap counts.
 *           H, V, D, the dynamic programming table
 *********************************************************************/
int
reverse_optimistic_table_weighted (char **A, char **B, int k, int l, int m, int n, 
		  int ***H, int ***V, int ***D) {
  
  int     i, j, x, y, p, q;
  int   a, b, c, d, oc, min, sub;
  int     len;
  char  **al;
  char  **Arev, **Brev;
  int **HR, **VR, **DR; 

  Arev = aa_malloc(k * sizeof(char *));
  for (i = 0; i < k; i++) { Arev[i] = aa_malloc((m+1) * sizeof(char)); }
  Brev = aa_malloc(l * sizeof(char *));
  for (i = 0; i < l; i++) { Brev[i] = aa_malloc((n+1) * sizeof(char)); }

  for (i = 0; i < k; i++) {
    for (j = 0; j < m; j++) {
      Arev[i][j] = A[i][m-j-1];
    }
    Arev[i][m] = '\0';
  }
  for (i = 0; i < l; i++) {
    for (j = 0; j < n; j++) {
      Brev[i][j] = B[i][n-j-1];
    }
    Brev[i][n] = '\0';
  }


  /* allocate and fill in boundary entries for dynamic programming table */
  *H = aa_malloc((m+1) * sizeof(int *));
  *V = aa_malloc((m+1) * sizeof(int *));
  *D = aa_malloc((m+1) * sizeof(int *));
  for (i = 0; i <= m; i++) {
    (*H)[i] = aa_malloc((n+1) * sizeof(int));
    (*V)[i] = aa_malloc((n+1) * sizeof(int));
    (*D)[i] = aa_malloc((n+1) * sizeof(int));
  }

  HR = aa_malloc((m+1) * sizeof(int *));
  VR = aa_malloc((m+1) * sizeof(int *));
  DR = aa_malloc((m+1) * sizeof(int *));
  for (i = 0; i <= m; i++) {
    HR[i] = aa_malloc((n+1) * sizeof(int));
    VR[i] = aa_malloc((n+1) * sizeof(int));
    DR[i] = aa_malloc((n+1) * sizeof(int));
  }


  for (i=0; i<=m; i++) {
  	for (j=0; j<=n; j++) {
	  HR[i][j] = VR[i][j] = DR[i][j] = 0;
  	}
  }

  //this gets unflipped after the DP table is filled
  flipFirstLastCharArrays();

  //fill in H[0][0] and V[0][0]
  for (p=0; p<k; p++) {
  	for (q=0; q<l; q++) {
  		if (Arev[p][0] == AA_EPS  && Brev[q][0] != AA_EPS) {
			HR[0][0] +=  aa_weights[p][k+q] * GET_GAMMA_A(p,0);
  		} else if (Arev[p][0] != AA_EPS  && Brev[q][0] == AA_EPS) {
	  		VR[0][0] += aa_weights[p][k+q] * GET_GAMMA_B(q,0);
  		}
  	}
  }
//  printf ("weighted: HR[0][0] = %d,   VR[0][0] = %d\n", HR[0][0], VR[0][0]);

  //fill in H[0][1], V[0][1] & D[0][1]
  for (p=0; p<k; p++) {
	 for (q=0; q<l; q++) {
  		if (Brev[q][0] != AA_EPS) {
			HR[0][1] += aa_weights[p][k+q] * (aa_lambda_terminal + aa_gamma_terminal );
  		}
/*  	This caused the boundary entries to be larger than in the unweighted table ... which seemed right to me.  
 * 		But it caused premature stopping of the table fill-in. Why was it wrong?
 		if (aa_firstCharsA[p] == 0) {//if posn 0 is the first non-gap in string A[i]) {
  			VR[0][1] += aa_weights[p][k+q] * GET_GAMMA_B(q,0);
  		}*/
 		if (aa_firstCharsA[p] == 0) {//if posn 0 is the first non-gap in string A[i]) {
 			int x = GET_GAMMA_B(q,0);
//			printf ( "Setting [0][1]: %d\n", x);
  			VR[0][1] += aa_weights[p][k+q] * x ; //GET_GAMMA_B(q,0);
  		}
	 }
  }
  DR[0][1] = HR[0][1];
  VR[0][1] += HR[0][1];


//  printf ("weighted: HR[0][1] = %d,   VR[0][1] = %d,   DR[0][1] = %d\n", HR[0][1], VR[0][1], DR[0][1]);


  //fill in H[1][0], V[1][0] & D[1][0]
  for (p=0; p<k; p++) {
	 for (q=0; q<l; q++) {
  		if (Arev[p][0] != AA_EPS) {
			VR[1][0] += aa_weights[p][k+q] * ( aa_lambda_terminal + aa_gamma_terminal ) ; // gap in B appears before the first column of B - must be terminal
  		}
/*  	This caused the boundary entries to be larger than in the unweighted table ... which seemed right to me.  
 * 		But it caused premature stopping of the table fill-in. Why was it wrong?
  		if (aa_firstCharsB[q] == 0) { //posn 0 is the first non-gap in string B[j]) {
  			HR[1][0] += aa_weights[p][k+q] * GET_GAMMA_A(p,0);
  		}
  		*/
  		if (aa_firstCharsB[q] == 0) { //posn 0 is the first non-gap in string B[j]) {
			int x = GET_GAMMA_A(p,0);
//			printf ( "Setting [1][0]: %d\n", x);
  			HR[1][0] += aa_weights[p][k+q] * x ; //GET_GAMMA_A(p,0);
  		}
	 }
  }
  DR[1][0] = VR[1][0];
  HR[1][0] += VR[1][0];

//  printf ("weighted: HR[1][0] = %d,   VR[1][0] = %d,   DR[1][0] = %d\n", HR[1][0], VR[1][0], DR[1][0]);

	  	  
  //fill in first row of DP table	  	  
  for (i = 2; i <= n; i++) {
  	HR[0][i] = HR[0][i-1];
	for (p=0; p<k; p++) {
		for (q=0; q<l; q++) {
	  		if (Brev[q][i-1] != AA_EPS) {
		    	HR[0][i] += aa_lambda_terminal * aa_weights[p][k+q];
	  		}
			if (aa_firstCharsB[q] == i-1) { //posn x-1 is the first non-gap in string B[j]) {
				HR[0][i] += aa_gamma_terminal * aa_weights[p][k+q];
			}
			if (aa_firstCharsA[p] == 0) { //posn 0 is the first non-gap in string A[j]) {
				VR[0][i] += aa_weights[p][k+q] * GET_GAMMA_B(k+q,i-1);
			}
		}
	}
    DR[0][i] = HR[0][i];
    VR[0][i] += HR[0][i];
        
  }
  
  for (i = 2; i <= m; i++) {
  	VR[i][0] = VR[i-1][0];
	for (p=0; p<k; p++) {
		for (q=0; q<l; q++) {
	  		if (Arev[p][i-1] != AA_EPS) {
		    	VR[i][0] += aa_lambda_terminal * aa_weights[p][k+q];
	  		}
			if (aa_firstCharsA[p] == i-1) { //posn x-1 is the first non-gap in string A[i]) {
				VR[i][0] += aa_gamma_terminal * aa_weights[p][k+q];
			}
			if (aa_firstCharsB[q] == 0) { //posn 0 is the first non-gap in string B[j]) {
				HR[i][0] +=  aa_weights[p][k+q] * GET_GAMMA_A(p, i-1);
			}
		}
	}
    DR[i][0] = VR[i][0];    
    HR[i][0] += VR[i][0];
  }


  
  /* fill in the dynamic programming table (row major order) */
  for (i = 1; i <= m; i++) {
    for (j = 1; j <= n; j++) {
		calcCost ( HR, VR, DR, Arev, Brev, AA_HORIZ, i, j, k, l, 0, &a, &b, &c, &d);
		HR[i][j] = d + minABC(a,b,c);
      	
		calcCost ( HR, VR, DR, Arev, Brev, AA_VERT, i, j, k, l, 0, &a, &b, &c, &d);
		VR[i][j] = d + minABC(a,b,c);

		calcCost ( HR, VR, DR, Arev, Brev, AA_DIAG, i, j, k, l, 0, &a, &b, &c, &d);
		DR[i][j] = d + minABC(a,b,c);
    }
  }
  flipFirstLastCharArrays(); // flip them back again


  a = HR[m][n];
  b = VR[m][n];
  c = DR[m][n];
  if (a < b) {   min = a; j = 1; }
  else {         min = b; j = 2; }
  if (c < min) { min = c; j = 3; }
  oc = min;

  /* construct the alignment (to calculate actual aa_cost)            */
  /* first pass to calculate length (for allocation), then recover */
  /* bug here for aa_gamma = 0, x can become negative for example */
  len = 0; x = m; y = n;
  while (x > 0 && y > 0) {
    len++;
    switch (j) {
    case 1: 
	  calcCost ( HR, VR, DR, Arev, Brev, AA_HORIZ, x, y, k, l, 0, &a, &b, &c, &d);
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      y--;
      break;

    case 2:
	  calcCost ( HR, VR, DR, Arev, Brev, AA_VERT, x, y, k, l, 0, &a, &b, &c, &d);
	  if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--;
      break;

    case 3:
	  calcCost ( HR, VR, DR, Arev, Brev, AA_DIAG, x, y, k, l, 0, &a, &b, &c, &d);
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--; y--;
      break;
    };
  }

  if (0 < x)      { len += x; }
  else if (0 < y) { len += y; }

  /* @@@ make routine to compute actual aa_cost without the actual alignment */
  al = aa_malloc((k+l) * sizeof(char *));
  for (i = 0; i < k+l; i++) { al[i] = aa_malloc((len+1) * sizeof(char)); }
  a = HR[m][n];
  b = VR[m][n];
  c = DR[m][n];
  if (a < b) { min = a; j = 1; } else { min = b; j = 2; }
  if (c < min) { min = c; j = 3; }
  sub = len; x = m; y = n;

  while (x > 0 && y > 0) {
//    printf ("%d,%d\n",x,y);
    switch (j) {
    case 1: 
      for (i = 0; i < k; i++) {	al[i][len-1] = AA_EPS; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = Brev[i][y-1]; }
	  calcCost ( HR, VR, DR, Arev, Brev, AA_HORIZ, x, y, k, l, 0, &a, &b, &c, &d);
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      y--;
      break;
    case 2:
      for (i = 0; i < k; i++) {	al[i][len-1] = Arev[i][x-1]; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = AA_EPS; }
	  calcCost ( HR, VR, DR, Arev, Brev, AA_VERT, x, y, k, l, 0, &a, &b, &c, &d);
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--;
      break; 
    case 3:
      for (i = 0; i < k; i++) {	al[i][len-1] = Arev[i][x-1]; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = Brev[i][y-1]; }
	  calcCost ( HR, VR, DR, Arev, Brev, AA_DIAG, x, y, k, l, 0, &a, &b, &c, &d);
      if (a < b) { min = a; j = 1; }
      else { min = b; j = 2; }
      if (c < min) { j = 3; }
      x--; y--;
      break;
    };
    len--;
  }

  if (0 < x)      { 
    while (0 < x) {
//      printf ("%d,%d\n",x,y);
      for (i = 0; i < k; i++) {	al[i][len-1] = Arev[i][x-1]; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = AA_EPS; }
      x--;
      len--;
    }
  }
  else if (0 < y) {
    while (0 < y) {
//      printf ("%d,%d\n",x,y);
      for (i = 0; i < k; i++) {	al[i][len-1] = AA_EPS; }
      for (i = 0; i < l; i++) {	al[i+k][len-1] = Brev[i][y-1]; }
      y--;
      len--;
    }
  }
  
  len = sub;
//  c = alignmentCost_special(al, al + k, k, l, len, len,1);
  c = alignmentCost(al, al + k, k, l, len, len);

  for (i = 0; i <= m; i++) {
    for (j = 0; j <=n; j++) {
        (*H)[i][j] = HR[m-i][n-j];
        (*V)[i][j] = VR[m-i][n-j];
        (*D)[i][j] = DR[m-i][n-j];			   				      
    }
  }

/*  printf("H\n");
  for (i = 0; i <= m; i++) {
    for (j = 0; j <=n; j++) {
        printf ("%d\t",(*H)[m-i][n-j]);
    }
    printf ("\n");
  }
  printf("V\n");
  for (i = 0; i <= m; i++) {
    for (j = 0; j <=n; j++) {
        printf ("%d\t",(*V)[m-i][n-j]);
    }
    printf ("\n");
  }
  printf("D\n");
  for (i = 0; i <= m; i++) {
    for (j = 0; j <=n; j++) {
        printf ("%d\t",(*D)[m-i][n-j]);
    }
    printf ("\n");
  }*/
  for (i = 0; i < k+l; i++) { aa_free(al[i], (len+1) * sizeof(char)); }
  aa_free(al, (k+l) * sizeof(char));
  for (i = 0; i < k; i++) { aa_free(Arev[i], (m+1) * sizeof(char)); }
  aa_free(Arev, k * sizeof(char *));
  for (i = 0; i < l; i++) { aa_free(Brev[i], (n+1) * sizeof(char)); }
  aa_free(Brev, l * sizeof(char *));
  for (i = 0; i <= m; i++) {
    aa_free(HR[i], (n+1) * sizeof(int));
    aa_free(VR[i], (n+1) * sizeof(int));
    aa_free(DR[i], (n+1) * sizeof(int));
  }
  aa_free(HR, (m+1) * sizeof(int *));
  aa_free(VR, (m+1) * sizeof(int *));
  aa_free(DR, (m+1) * sizeof(int *));

  return c;
}


inline int
minABC (int a, int b, int c) {
	return (a<b)?( (a<c)?a:c ) : ( (b<c)?b:c );	
}
